@@ -6,6 +6,8 @@
 // Must be in a non-cgo-using package so that
 // the go command compiles this file with 6c, not gcc.
 
+// +build gc
+
 typedef char bool;
 
 bool runtime·lockedOSThread(void);
@@ -21,3 +23,10 @@ void
 	b = runtime·lockedOSThread();
 	FLUSH(&b);
 }
+
+// This is what a cgo-compiled stub declaration looks like.
+void
+·Issue7695(struct{void *y[8*sizeof(void*)];}p)
+{
+	USED(p);
+}
