@@ -8,6 +8,12 @@
 
 int	thechar	= '8';
 char*	thestring	= "386";
+LinkArch*	thelinkarch = &link386;
+
+void
+linkarchinit(void)
+{
+}
 
 vlong MAXWIDTH = (1LL<<32) - 1;
 
@@ -27,6 +33,8 @@ void
 betypeinit(void)
 {
 	widthptr = 4;
+	widthint = 4;
+	widthreg = 4;
 
 	zprog.link = P;
 	zprog.as = AGOK;
@@ -35,5 +43,5 @@ betypeinit(void)
 	zprog.from.scale = 0;
 	zprog.to = zprog.from;
 
-	listinit();
+	listinit8();
 }
