@@ -14,25 +14,28 @@ static struct {
 	// is converted by bisonerrors into the yystate and yychar caused
 	// by that token list.
 
-	221, ',',
+	222, ',',
 	"unexpected comma during import block",
 
-	377, ';',
-	"unexpected semicolon or newline before {",
+	32, ';',
+	"missing import path; require quoted string",
 
-	398, ';',
-	"unexpected semicolon or newline before {",
+	378, ';',
+	"missing { after if clause",
 
-	237, ';',
-	"unexpected semicolon or newline before {",
+	399, ';',
+	"missing { after switch clause",
 
-	475, LBODY,
-	"unexpected semicolon or newline before {",
+	238, ';',
+	"missing { after for clause",
+
+	476, LBODY,
+	"missing { after for clause",
 
 	22, '{',
 	"unexpected semicolon or newline before {",
 
-	144, ';',
+	145, ';',
 	"unexpected semicolon or newline in type declaration",
 
 	37, '}',
@@ -44,33 +47,33 @@ static struct {
 	37, ',',
 	"unexpected comma in channel type",
 
-	438, LELSE,
+	439, LELSE,
 	"unexpected semicolon or newline before else",
 
-	257, ',',
+	258, ',',
 	"name list not allowed in interface type",
 
-	237, LVAR,
+	238, LVAR,
 	"var declaration not allowed in for initializer",
 
 	65, '{',
 	"unexpected { at end of statement",
 
-	376, '{',
+	377, '{',
 	"unexpected { at end of statement",
 	
-	125, ';',
+	126, ';',
 	"argument to go/defer must be function call",
 	
-	425, ';',
+	426, ';',
 	"need trailing comma before newline in composite literal",
 	
-	436, ';',
+	437, ';',
 	"need trailing comma before newline in composite literal",
 	
-	112, LNAME,
+	113, LNAME,
 	"nested func not allowed",
 
-	642, ';',
+	645, ';',
 	"else must be followed by if or statement block"
 };
