@@ -1,4 +1,4 @@
-// +build 386 arm
+// +build 386 amd64p32 arm
 // errorcheck
 
 // Copyright 2011 The Go Authors.  All rights reserved.
@@ -9,7 +9,7 @@
 
 package main
 func main() {
-	var arr [1000200030]int   // ERROR "type .* too large"
+	var arr [1000200030]int   // GC_ERROR "type .* too large"
 	arr_bkup := arr
 	_ = arr_bkup
 }
